{
@abstract(defines all items that can appear within a Pascal unit's interface)
@created(11 Mar 1999)
@lastmod(10 Feb 2000)
@author(Marco Schmidt (marcoschmidt@geocities.com))

For each type, variable, class etc. (short: item) that may appear in a Pascal
source code file and can thus be taken into the documentation, this unit
provides an object type which will store name, unit, description and more
on this item.
}
unit items;

interface

uses
  objects, texts;

const
  { indicates field or method is public (default state) }
  STATE_PUBLIC = 0;
  { indicates field or method is protected }
  STATE_PROTECTED = 1;
  { indicates field or method is private }
  STATE_PRIVATE = 2;
  { indicates field or method is published }
  STATE_PUBLISHED = 3;

type
  { pointer to @link(TItemCollection) }
  PItemCollection = ^TItemCollection;
  { pointer to @link(TMethodCollection) }
  PMethodCollection = ^TMethodCollection;
  { pointer to @link(TPropertyCollection) }
  PPropertyCollection = ^TPropertyCollection;
  { pointer to @link(TMethod) }
  PMethod = ^TMethod;
  { pointer to @link(TProperty) }
  PProperty = ^TProperty;
  { pointer to @link(TUnit) }
  PUnit = ^TUnit;
  { pointer to @link(TObjectInfo) }
  PCIO = ^TCIO;
  { pointer to @link(TItem) }
  PItem = ^TItem;
  { basic linkable item in pasdoc hierarchy }
  TItem = object(TObject)
    { unique number of this item }
    AnchorNumber: LongInt;
    { list of strings, each representing one author of this item }
    Authors: PStringCollection;
    { if assigned, contains string with date of creation }
    Created: PString;
    { description of this item, a single sentence }
    Description: PText;
    { more detailed description of this item, mostly more than one
      sentence }
    DetailedDescription: PText;
    { a full link that should be enough to link this item from anywhere else }
    FullLink: String[127];
    { if assigned, contains string with date of last modification }
    LastMod: PString;
    { if this item is part of an object or class, the corresponding
      info object is stored here, nil otherwise }
    MyObject: PCIO;
    { pointer to unit this item belongs to }
    MyUnit: PUnit;
    { name of the item }
    Name: String;
	{ One of the STATE_xxx constants, determines access rights
	  (public, private, etc.). }
	State: Byte;
    { }
    destructor Done; virtual;
    function FindItem(ItemName: String): PItem; virtual;
    { }
    function FindName(S1, S2, S3: String; N: Integer): PItem; virtual;
    { Returns DetailedDescription if available, otherwise Description,
      otherwise nil. }
    function GetDescription: PText;
    { Searches for an abstract tag within the Description field of
      this item. If one is found, Description is copied to DetailedDescription
      and the abstract tag becomes the new Description. This procedure
      should be called after the dates (created and lastmod) and the
      author tags have been handled, as they are searched in Description. }
    { }
    procedure HandleAbstractTag;
    { }
    procedure HandleAuthorTags;
    { }
    procedure HandleCreatedTag;
    { }
    procedure HandleLastmodTag;
    { Returns true if there is a detailled or a normal description available. }
    function HasDescription: Boolean;
    { Inserts an item into a collection.
	  Creates collection if it does not exist already. }
    procedure InsertItem(Item: PItem; var C: PItemCollection);
    { inserts a method into a collection; creates collection if necessary }
    procedure InsertMethod(Method: PMethod; var C: PMethodCollection);
    { inserts a property into a collection; creates collection if necessary }
    procedure InsertProperty(Prop: PProperty; Var C: PPropertyCollection);
  end;
  { extends @link(TItem) to store method and function-/procedure-specific
    information }
  TMethod = object(TItem)
    { full declaration, including parameter list and procedural directives }
    FullDeclaration: String;
    { 0=constructor 1=destructor 2=function or procedure }
    What: Byte;
  end;
  TProperty = object(TItem)
    { full declaration, including read/write and storage specifiers }
    FullDeclaration: String;
    { contains the optional index declaration, including brackets }
    IndexDecl : String;
    { contains the type of the property }
    Proptype : String;
    { read specifier }
    Reader: String;
    { write specifier }
    Writer: String;
    { true if the property is the default property }
    Default : Boolean;
    { keeps default value specifier }
    DefaultID : String;
    { true if Nodefault property }
    NoDefault : Boolean;
    { keeps Stored specifier }
    StoredId : String;
    { Get a string that gives a short declaration, just giving the name and 
      the type, optionally the index. }
  {  function GetDeclaration : String;}
    { Get a string that describes the specifiers such as read-only, 
      write-only etc. If Detailed is true, then the values are 
      specified too. }
  {  function GetSpecifiers(Detailed : Boolean) : String;}
  end;
  { enumeration type to determine type of TObjectInfo item: class,
    interface or object }
  TCIOType = (CIO_CLASS, CIO_INTERFACE, CIO_OBJECT);
  { Extends @link(TItem) to store all items in a class / an object, e.g.
    fields. }
  TCIO = object(TItem)
    { name of the ancestor class / object }
    Ancestors: PStringCollection;
    { list of all fields }
    Fields: PItemCollection;
    { list of all methods }
    Methods: PMethodCollection;
    { determines if this is a class, an interface or an object }
    MyType: TCIOType;
    { name of documentation output file (if each class / object gets
      its own file, that's the case for HTML, but not for TeX) }
    OutputFileName: String;
    { list of properties }
    Properties : PPropertyCollection;
    { Simply returns the result of a call to @link(FindFieldMethodProperty). }
    function FindItem(ItemName: String): PItem; virtual;
    { If this class (or interface or object) contains a field, method or
      property with the name of ItemName, the corresponding item pointer is
      returned. }
    function FindFieldMethodProperty(ItemName: String): PItem;
    {function GetType: Char;}
  end;
  { extends @link(TItem) to store anything about a unit, its constants,
    types etc.; also provides methods for parsing a complete unit }
  TUnit = object(TItem)
    { list of classes and objects defined in this unit }
    CIO: PItemCollection;
    { list of constants defined in this unit }
    Constants: PItemCollection;
    { list of functions and procedures defined in this unit }
    FuncsProcs: PMethodCollection;
    { name of documentation output file
	  THIS SHOULD NOT BE HERE! }
    OutputFileName: String;
    { the names of all units mentioned in a uses clause in the interface
      section of this unit }
    UsesUnits: PStringCollection;
    { list of classes and objects defined in this unit }
    SourceFileName: String;
    { list of types defined in this unit }
    Types: PItemCollection;
    { list of variables defined in this unit }
    Variables: PItemCollection;
    { create this object with given input stream and file name }
    constructor Init;
    { dispose of all dynamically allocated memory in this object }
    destructor Done; virtual;
    procedure AddCIO(I: PCIO);
    procedure AddConstant(I: PItem);
    procedure AddType(I: PItem);
	procedure AddVariable(I: PItem);
    function FindFieldMethodProperty(S1, S2: String): PItem;
    function FindItem(ItemName: String): PItem; virtual;
    procedure RemovePrivateItems;
  end;
  { extends @link(TSortedCollection) to store a list of items }
  TItemCollection = object(TSortedCollection)
    { Moves all elements from C to this object, but does not
      dispose of the - empty - C }
    procedure AddItems(var C: PItemCollection);
    { Compares two @link(TItem) objects by their @link(TItem.Name) fields. }
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    { Counts classes, interfaces and objects within this collection. }
    procedure CountCIO(var C, I, O: LongInt);
    { Compares each element's name field with Name and returns the item on
      success, nil otherwise.
      Name's case is not regarded. }
    function FindName(Name: String): PItem;
    { Inserts all items of C into this collection.
      Disposes C and sets it to nil. }
    procedure InsertItems(var C: PItemCollection);
    { Checks each element's State field and removes all elements with a value
      of STATE_PRIVATE. }
    procedure RemovePrivateItems;
  end;
  { extends @link(TItemCollection) by defining a different compare function: two
    methods are first compared by their @link(TMethod.What) field, which determines
    whether they are constructor, destructor or function / procedure }
  TMethodCollection = object(TItemCollection)
    { compares two method objects (it is assumed Key1 and
      Key2 are of type PMethod) }
    function Compare(Key1, Key2: Pointer): Integer; virtual;
  end;
  { a TPropertyCollection holds a collection of properties. It introduces no
    new methods when compared to @link(TITemCollection), but this may be
    implemented in a later stage }
  TPropertyCollection = Object(TItemCollection)
  end;

{ If C is assigned and C^.Count is zero, the collection's destructor is
  called and C is set to nil.
  If C is nil, nothing happens. }
procedure FreeEmptyCollection(var C: PItemCollection);

implementation

uses
  arrays, msg, numbers;

procedure FreeEmptyCollection(var C: PItemCollection);
begin
  if (not Assigned(C))
  then exit;
  if (C^.Count < 1)
  then
    begin
      Dispose(C, Done);
      C := nil;
    end;
end;

{ TItem }

destructor TItem.Done;
begin
  if Assigned(Authors)
  then Dispose(Authors, Done);
  if Assigned(Created)
  then DisposeStr(Created);
  if Assigned(Description)
  then Dispose(Description, Done);
  if Assigned(DetailedDescription)
  then Dispose(DetailedDescription, Done);
  if Assigned(LastMod)
  then DisposeStr(LastMod);
  inherited Done;
end;

function TItem.FindItem(ItemName: String): PItem;
begin
  FindItem := nil;
end;

function TItem.FindName(S1, S2, S3: String; N: Integer): PItem;
var
  P: PItem;
begin
  FindName := nil;
  if (N <> 0) and (N <> 1)
  then exit;
  { is S1.S2.S3 a link to the unit this item is stored in? }
  if (N = 0) and Assigned(MyUnit) and (CompareStringsCI(S1, MyUnit^.Name) = 0)
  then
    begin
      FindName := MyUnit;
      exit;
    end;
  if Assigned(MyObject)
  then
    begin { this item is a method or field }
      if (N = 0)
      then
        begin
          P := MyObject^.FindFieldMethodProperty(S1);
          if Assigned(P)
          then
            begin
              FindName := P;
              exit;
            end;
         end
       else
         begin { N must be 1 }
           if (CompareStringsCI(MyObject^.Name, S1) = 0)
           then
             begin
                P := MyObject^.FindFieldMethodProperty(S2);
                if Assigned(P)
                then
                  begin
                    FindName := P;
                    exit;
                  end;
             end;
         end;
    end;
  if Assigned(MyUnit) and (N = 0)
  then
    begin
    end;
  if Assigned(MyUnit) and (N = 0)
  then
    begin
      P := MyUnit^.FindItem(S1);
      if Assigned(P)
      then FindName := P;
    end;
end;

function TItem.GetDescription: PText;
begin
  if Assigned(DetailedDescription)
  then GetDescription := DetailedDescription
  else
    begin
      if Assigned(Description)
      then GetDescription := Description
      else GetDescription := nil;
    end;
end;

procedure TItem.HandleAbstractTag;
var
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  S: String;
  P: PText;
begin
  Offs1 := 0;
  if (not Assigned(Description)) or
     (not Description^.FindTag('ABSTRACT', Offs1, Offs2, Offs3))
  then exit;
  Description^.ExtractTag(Offs1, Offs2, Offs3, S);
  P := New(PText, Init);
  P^.AppendString(S);
  DetailedDescription := Description;
  Description := P;
end;

procedure TItem.HandleAuthorTags;
var
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  P: PString;
  S: String;
begin
  Authors := nil;
  if (not Assigned(Description))
  then exit;
  Offs1 := 0;
    { we could have more than one author, so repeat until we have all }
  repeat
    if Description^.FindTag('AUTHOR', Offs1, Offs2, Offs3) then
    begin
      { we found one, remove it from the description and add it
        to the author list }
      Description^.ExtractTag(Offs1, Offs2, Offs3, S);
      if (Length(S) < 0)
      then Break;
      if (not Assigned(Authors))
      then Authors := New(PStringCollection, Init(16, 16));
      P := NewStr(S);
      if (not Assigned(P)) then
      begin
      	PrintLn(1, 'ERROR: Could not allocate memory for string "' + S + '"');
      	exit;
      end
      else Authors^.Insert(P)
    end
    else exit; { no more authors in Description -> leave }
  until false;
end;

procedure TItem.HandleCreatedTag;
var
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  S: String;
begin
  Created := nil;
  if (not Assigned(Description))
  then exit;
  Offs1 := 0;
  if (not Description^.FindTag('CREATED', Offs1, Offs2, Offs3))
  then exit;
  Description^.ExtractTag(Offs1, Offs2, Offs3, S);
  Created := NewStr(S);
end;

procedure TItem.HandleLastmodTag;
var
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  S: String;
begin
  LastMod := nil;
  if (not Assigned(Description))
  then exit;
  Offs1 := 0;
  if (not Description^.FindTag('LASTMOD', Offs1, Offs2, Offs3))
  then exit;
  Description^.ExtractTag(Offs1, Offs2, Offs3, S);
  LastMod := NewStr(S);
end;

function TItem.HasDescription: Boolean;
begin
  HasDescription := (Assigned(Description) or Assigned(DetailedDescription));
end;

procedure TItem.InsertItem(Item: PItem; var C: PItemCollection);
begin
  if (not Assigned(Item))
  then exit;
  if (not Assigned(C))
  then C := New(PItemCollection, Init(16, 16));
  C^.Insert(Item);
end;

procedure TItem.InsertMethod(Method: PMethod; var C: PMethodCollection);
begin
  if (not Assigned(Method))
  then exit;
  if (not Assigned(C))
  then C := New(PMethodCollection, Init(16, 16));
  C^.Insert(Method);
end;

procedure TItem.InsertProperty(Prop : PProperty; Var C: PPropertyCollection);

begin
        if (not assigned(Prop)) then exit;
        if (not Assigned(C)) then
          C:=New(PPropertyCollection,Init(16,16));
        C^.Insert(Prop); 
end;

{ TItemCollection }

procedure TItemCollection.AddItems(var C: PItemCollection);
var
  I: LongInt;
  P: Pointer;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  for I := 0 to C^.Count - 1 do
  begin
    P := C^.At(I);
    Insert(P);
  end;
  C^.DeleteAll;
end;

function TItemCollection.Compare(Key1, Key2: Pointer): Integer;
var
  U1: String;
  U2: String;
begin
  StringToUpper(PItem(Key1)^.Name, U1);
  StringToUpper(PItem(Key2)^.Name, U2);
  if (U1 = U2)
  then Compare := 0
  else
    begin
      if (U1 < U2)
      then Compare := -1
      else Compare := 1;
    end;
end;

procedure TItemCollection.CountCIO(var C, I, O: LongInt);
var
  J: LongInt;
  P: PCIO;
begin
  C := 0;
  I := 0;
  O := 0;
  if (Count < 1)
  then exit;
  J := 0;
  repeat
    P := At(J);
    case P^.MyType of
      CIO_CLASS: Inc(C);
      CIO_INTERFACE: Inc(I);
      CIO_OBJECT: Inc(O);
    end;
    Inc(J);
  until (J = Count);
end;

function TItemCollection.FindName(Name: String): PItem;
var
  I: Integer;
  Item: PItem;
begin
  if (Count < 1) then
  begin
    FindName := nil;
    exit;
  end;
  for I := 0 to Count - 1 do
  begin
    Item := At(I);
    if Assigned(Item) and
       (Arrays.CompareStringsCI(Name, Item^.Name) = 0) then
    begin
      FindName := Item;
      exit;
    end;
  end;
  FindName := nil;
end;

procedure TItemCollection.InsertItems(var C: PItemCollection);
var
  I: LongInt;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  I := C^.Count - 1;
  repeat
    Insert(C^.At(I));
    C^.AtDelete(I);
    Dec(I);
  until (I = -1);
end;

procedure TItemCollection.RemovePrivateItems;
var
  I: LongInt;
  Item: PItem;
begin
  I := 0;
  while (I < Count) do
  begin
    Item := At(I);
    if Assigned(Item) and (Item^.State = STATE_PRIVATE) then
    begin
      AtFree(I);
    end
    else Inc(I);
  end;
end;


{
function TItemCollection.SplitLink(S: String; var S1, S2, S3: String; var N: LongInt): Boolean;
var
  I: LongInt;
begin
  SplitLink := false;
  S1 := '';
  S2 := '';
  S3 := '';
  N := 0;
  I := 1;
  while (I <= Length(S)) and (not (S[I] in IdentifierStart)) do
    Inc(I);
  if (I > Length(S))
  then exit;
  if (I > 1)
  then System.Delete(S, 1, I - 1);
  I := 2;
  while (I <= Length(S)) and (S[I] in IdentifierOther) do
    Inc(I);
  System.Delete(S, I - 1, Length(S) - I + 1);
  I := System.Pos('.', S);
  SplitLink := true;
  if (I = 0)
  then
    begin
      S1 := S;
      N := 1;
      exit;
    end;
  S1 := System.Copy(S, 1, I - 1);
  System.Delete(S, 1, I);
  I := System.Pos('.', S);
end;
}

function TCIO.FindItem(ItemName: String): PItem;
begin
  FindItem := FindFieldMethodProperty(ItemName);
end;

function TCIO.FindFieldMethodProperty(ItemName: String): PItem;
var
  C: PItemCollection;
  I: Integer;
  Item: PItem;
begin
  for I := 0 to 2 do
  begin
    case I of
      0: C := Fields;
      1: C := Methods;
      2: C := Properties;
      else C := nil;
    end;
    if Assigned(C) then
    begin
      Item := C^.FindName(ItemName);
      if Assigned(Item) then
      begin
        FindFieldMethodProperty := Item;
        exit;
      end;
    end;
  end;
  FindFieldMethodProperty := nil;
end;
{
function TCIO.GetType: Char;
begin
  case MyType of
    CIO_CLASS: GetType := 'C';
    CIO_INTERFACE: GetType := 'I';
    CIO_OBJECT: GetType := 'O';
    else GetType := '?';
  end;
end;
}
{ TMethodCollection }

function TMethodCollection.Compare(Key1, Key2: Pointer): Integer;
var
  P1: PMethod;
  P2: PMethod;
  R: Integer;
begin
  P1 := PMethod(Key1);
  P2 := PMethod(Key2);
  { compare 'method type', order is constructor > destructor > function, procedure }
  R := Numbers.CompareUInt8(P1^.What, P2^.What);
  if (R <> 0)
  then
    begin
      Compare := R;
      exit;
    end;
  { if 'method type' is equal, compare names }
  Compare := Arrays.CompareStringsCI(P1^.Name, P2^.Name);
end;

{ TUnit }

constructor TUnit.Init;
begin
  inherited Init;
{  Stream := New(PUnitStream, Init(InputStream));}
{  SourceFileName := FileName;}
  Types := nil;
  Constants := nil;
  Variables := nil;
  FuncsProcs := nil;
  CIO := nil;
end;

destructor TUnit.Done;
begin
{  if Assigned(Stream)
  then Dispose(Stream, Done);}
  if Assigned(Types)
  then Dispose(Types, Done);
  inherited Done;
end;

procedure TUnit.AddCIO(I: PCIO);
begin
  if (not Assigned(CIO))
  then
    begin
      CIO := New(PItemCollection, Init(16, 16));
      if (not Assigned(CIO))
      then exit;
    end;
  CIO^.Insert(I);
end;

procedure TUnit.AddConstant(I: PItem);
begin
  if (not Assigned(Constants)) then
    begin
      Constants := New(PItemCollection, Init(16, 16));
      if (not Assigned(Constants)) then
      begin
        PrintLn(1, 'Error - could not create list of constants.');
        Halt(1);
      end;
    end;
  Constants^.Insert(I);
end;

procedure TUnit.AddType(I: PItem);
begin
  if (not Assigned(Types)) then
  begin
    Types := New(PItemCollection, Init(16, 16));
    if (not Assigned(Types)) then
    begin
      PrintLn(1, 'Error - could not create list of types.');
      Halt(1);
    end;
  end;
  Types^.Insert(I);
end;

procedure TUnit.AddVariable(I: PItem);
begin
  if (not Assigned(Variables))
  then
    begin
      Variables := New(PItemCollection, Init(16, 16));
      if (not Assigned(Variables))
      then exit;
    end;
  Variables^.Insert(I);
end;

function TUnit.FindFieldMethodProperty(S1, S2: String): PItem;
var
  PI: PItem;
  PO: PCIO;
begin
  FindFieldMethodProperty := nil;
  if (not Assigned(CIO))
  then exit;
  PO := PCIO(CIO^.FindName(S1));
  if Assigned(PO) then
  begin
    PI := PO^.FindFieldMethodProperty(S2);
    if Assigned(PI)
    then FindFieldMethodProperty := PI;
  end;
end;

function TUnit.FindItem(ItemName: String): PItem;
var
  Item: PItem;
begin
  FindItem := nil;
    Item := nil;
  if (CompareStringsCI(Name, ItemName) = 0)
  then
    begin
      FindItem := @Self;
      exit;
    end;
  if Assigned(Constants)
  then Item := Constants^.FindName(ItemName);
  if Assigned(Item)
  then
    begin
      FindItem := Item;
      exit;
    end;
  if Assigned(Types)
  then Item := Types^.FindName(ItemName);
  if Assigned(Item)
  then
    begin
      FindItem := Item;
      exit;
    end;
  if Assigned(Variables)
  then Item := Variables^.FindName(ItemName);
  if Assigned(Item)
  then
    begin
      FindItem := Item;
      exit;
    end;
  if Assigned(FuncsProcs)
  then Item := FuncsProcs^.FindName(ItemName);
  if Assigned(Item)
  then
    begin
      FindItem := Item;
      exit;
    end;
  if Assigned(CIO)
  then Item := CIO^.FindName(ItemName);
  if Assigned(Item)
  then
    begin
      FindItem := Item;
      exit;
    end;
end;

procedure TUnit.RemovePrivateItems;
var
  I: LongInt;
  P: PCIO;
begin
{  WriteLn('Removing private items of unit ', Name);}
  if (not Assigned(CIO))
  then exit;
  I := 0;
  while (I < CIO^.Count) do
  begin
    P := PCIO(CIO^.At(I));

    if Assigned(P^.Fields)
    then P^.Fields^.RemovePrivateItems;
    FreeEmptyCollection(P^.Fields);

    if Assigned(P^.Methods)
    then P^.Methods^.RemovePrivateItems;
    FreeEmptyCollection(PItemCollection(P^.Methods));

    if Assigned(P^.Properties)
    then P^.Properties^.RemovePrivateItems;
    FreeEmptyCollection(PItemCollection(P^.Properties));

    Inc(I);
  end;
end;

{ TProperty }

{Function TProperty.GetDeclaration : String;

Var T : String;

begin
  T:='Property '+Name;
  If Length(IndexName)<>0 Then
  T:=T+'['+IndexName+':'+IndexType+']';
  T:=T+ ' : '+Proptype+';';
  GetDeclaration:=T;
end;

Function TProperty.GetSpecifiers (Detailed : Boolean) : String;

Var S : String;

Procedure addto (Const T : String);

begin
  If Length(S)>0 then
    S:=S+', '+T
  else
    S:=T;
end;

begin
  S:='';
  If Length(Reader)=0 Then
    begin
    If Not detailed then
     AddTo('Write Only')
    end
  else 
    If Detailed then 
       AddTo('Read '+Reader);
  if Length(Writer)=0 Then
    begin
    If not detailed then
      AddTo('Read Only')
    end
  else
    If Detailed then
      AddTo('Write '+Writer);
  If Default then 
    AddTo('Default property');
  If NoDefault then
    AddTo('NoDefault');
  If Length(DefaultID)<>0 then 
  If Not Detailed then
    AddTo('Has default')
  else
    AddTo('default '+DefaultId);
  If Length (StoredID)<>0 then 
    If Detailed then
      addTo('Stored '+StoredId)
    else
      AddTo ('Is stored');
  If S<>'' then 
    S:='('+S+')';
  GetSpecifiers:=S;
end;
}

{ unit initialization }

begin
end.