{
@abstract(provides simplified Pascal scanner)
@created(24 Feb 1999)
@lastmod(6 Apr 2000)
@author(Marco Schmidt (marcoschmidt@geocities.com))

The scanner object @link(TScanner) returns tokens from a Pascal language
character input stream.
It uses the @link(Tokenizi) unit to get tokens, regarding conditional
directives that might lead to including another files or will add or
delete conditional directives.
So, this scanner is a combined tokenizer and pre-processor.
}

unit scanning;

interface

uses
  objects, streams, tokenizi;

const
  { maximum number of streams we can recurse into; first one is the unit
    stream, any other stream an include file; current value is 32, increase
    this if you have more include files recursively including others }
  MAX_TOKENIZERS = 32;

type
  { subrange type that has the 26 lower case letters from a to z }
  TLowerCaseLetter = 'a' .. 'z';
  { an array of boolean values, index type is @link(TLowerCaseLetter) }
  TSwitchOptions = Array[TLowerCaseLetter] of Boolean;
  { a pointer to a @link(TScanner) object }
  PScanner = ^TScanner;
  { This class scans one unit using one or more @link(TTokenizer) objects
    to scan the unit and all nested include files.
  }
  TScanner = object(TObject)
    BufferedToken: PToken;
    CT: Integer;
    DirectiveLevel: Integer;
    Directives: PStringCollection;
    ErrorMessage: String;
    IncludeFilePaths: PCollection;
    IsTokenBuffered: Boolean;
    SwitchOptions: TSwitchOptions;
    Tokenizers: Array[0..MAX_TOKENIZERS - 1] of PTokenizer;
    { Creates a TScanner object that scans the given input stream. }
    constructor Init(S: PInputStream);
    destructor Done; virtual;
    { Adds parameter string to the list of directives.
      Returns true on success, false on failure (e.g. in case that there
      wasn't enough memory. }
    function AddDirective(N: String): Boolean;
    { Adds all directives in the parameter string collection by calling
      @link(AddDirective) for each of the strings in that collection. }
    function AddDirectives(DL: PStringCollection): Boolean;
    { Gets next token and throws it away. }
    procedure ConsumeToken;
    { Removes directive N from the internal list of directives.
      If N was not in that list, nothing is done. }
    procedure DeleteDirective(N: String);
    { Returns the name of the file that is currently processed and the line
      number. Good for meaningful error messages. }
	function GetStreamInfo: String;
	{ Returns next token as parameter. Returns true on success, false on error. }
    function GetToken(var T: PToken): Boolean;
    { Returns if a given directive N is defined at the moment. }
    function IsDirectiveDefined(N: String): Boolean;
    function IsSwitchDefined(N: String): Boolean;
    function OpenInclude(N: String): Boolean;
    function PeekToken(var T: PToken): Boolean;
    function SkipUntilElseOrEndif(var FoundElse: Boolean): Boolean;
    procedure UnGetToken(T: PToken);
    procedure UpdateSwitchOptions(O: String);
  end;

implementation

uses
  arrays, filestre, msg, texts;

type
  { all directives a scanner is going to regard }
  TDirectiveType = (DT_DEFINE, DT_ELSE, DT_ENDIF, DT_IFDEF, DT_IFNDEF,
    DT_IFOPT, DT_INCLUDE_FILE, DT_UNDEF, DT_UNKNOWN);

const
  NUM_DIRECTIVES = 8;
  DirectiveNames: Array[0..NUM_DIRECTIVES - 1] of String[31] =
    ('DEFINE', 'ELSE', 'ENDIF', 'IFDEF', 'IFNDEF', 'IFOPT', 'I', 'UNDEF');

{ this function recognizes only those directives we'll need for the scanner }
function SplitDirective(T: PText; var DirectiveName, Params: String): Boolean;
var
  I: LongInt;
begin
  SplitDirective := false;
  I := 0;
  DirectiveName := '';
  Params := '';
  { find dollar sign }
  while (I < T^.Content) and (T^.Data^[I] <> '$') do
    Inc(I);
  if (I = T^.Content)
  then exit;
  Inc(I);
  { get directive name }
  while (I < T^.Content) and (T^.Data^[I] <> ' ') and
    (T^.Data^[I] <> '*') and (T^.Data^[I] <> '}') do
  begin
    DirectiveName := DirectiveName + Upcase(T^.Data^[I]);
    Inc(I);
  end;
  SplitDirective := true;
  if (T^.Data^[I] = '*') or (T^.Data^[I] = '}')
  then exit;
  { skip spaces }
  while (I < T^.Content) and (T^.Data^[I] = ' ') do
    Inc(I);
  if (I = T^.Content)
  then exit;
  { get parameters - no conversion to uppercase here, it could be an include
    file name whose name need not be changed (platform.inc <> PLATFORM.INC) }
  while (I < T^.Content) and (T^.Data^[I] <> '*') and
    (T^.Data^[I] <> '}') do
  begin
    Params := Params + T^.Data^[I];
    Inc(I);
  end;
end;

function IdentifyDirective(T: PText; var DT: TDirectiveType; var P: String): Boolean;
var
  I: Integer;
  S: String;
begin
  IdentifyDirective := false;
  if (not SplitDirective(T, S, P))
  then exit;
  StringToUpper(S, S);
  for I := 0 to NUM_DIRECTIVES - 1 do
  begin
    if (S = DirectiveNames[I])
    then
      begin
        DT := TDirectiveType(I);
        IdentifyDirective := true;
        exit;
      end;
  end;
end;

{ TScanner }

constructor TScanner.Init(S: PInputStream);
var
  C: TLowerCaseLetter;
begin
  inherited Init;
  CT := 0;
  DirectiveLevel := 0;
  Directives := nil;
  ErrorMessage := '';
  IsTokenBuffered := false;
  for C := Low(TLowerCaseLetter) to High(TLowerCaseLetter) do
    SwitchOptions[C] := false;
  Tokenizers[0] := New(PTokenizer, Init(S));
  if (not Assigned(Tokenizers[0]))
  then Fail;
end;

destructor TScanner.Done;
var
  I: Integer;
begin
  I := 0;
  while (I <= CT) do
  begin
    if Assigned(Tokenizers[I])
    then Dispose(Tokenizers[I], Done);
    Inc(I);
  end;
  inherited Done;
end;

function TScanner.AddDirective(N: String): Boolean;
var
  S: PString;
begin
  StringToUpper(N, N);
  if (not IsDirectiveDefined(N))
  then
    begin
      AddDirective := false;
      if (not Assigned(Directives)) then
      begin
        Directives := New(PStringCollection, Init(16, 16));
        if (not Assigned(Directives))
        then exit;
      end;
      S := NewStr(N);
      if (not Assigned(S))
      then exit;
      Directives^.Insert(S);
      {WriteLn('added directive "' + N + '"');}
      AddDirective := true;
    end
  else
    begin
      {WriteLn('directive "' + N + '" already defined');}
	  AddDirective := true;
    end;
end;

function TScanner.AddDirectives(DL: PStringCollection): Boolean;
var
  S: PString;
  I: LongInt;
begin
  AddDirectives := true;
  if (not Assigned(DL)) or (DL^.Count < 1)
  then exit;
  for I := 0 to DL^.Count - 1 do
  begin
    S := DL^.At(I);
    if Assigned(S) then
	begin
	  if (not AddDirective(S^)) then
      begin
        AddDirectives := false;
        exit;
      end;
    end;
  end;
end;

procedure TScanner.ConsumeToken;
begin
  IsTokenBuffered := false;
end;

procedure TScanner.DeleteDirective(N: String);
var
  I: LongInt;
begin
  if (not Assigned(Directives))
  then exit;
  StringToUpper(N, N);
  I := Directives^.IndexOf(@N);
  if (I <> -1) then
  begin
    Directives^.AtFree(I);
    {WriteLn('deldir: deleted directive "' + N + '"');}
  end
  else {WriteLn('deldir: directive "' + N + '" is not defined')};
end;

function TScanner.GetStreamInfo: String;
begin
  if (CT >= 0)
  then GetStreamInfo := Tokenizers[CT]^.GetStreamInfo
  else GetStreamInfo := '';
end;

function TScanner.GetToken(var T: PToken): Boolean;
var
  DT: TDirectiveType;
  Finished: Boolean;
  FoundElse: Boolean;
  P: String;
begin
  if IsTokenBuffered then
  begin
  	{ we have a token buffered, we'll return this one }
    T := BufferedToken;
    IsTokenBuffered := false;
    GetToken := true;
    exit;
  end;
  GetToken := false;
  Finished := false;
  repeat
    { check if we have a tokenizer left }
    if (CT = -1) then
    begin
      PrintLn(1, 'Error: End of stream reached while trying to get next token.');
      exit;
    end;
    if Tokenizers[CT]^.HasData then
    begin
      { get next token from tokenizer }
      if (not Tokenizers[CT]^.GetToken(T)) then
      begin
        {writeln('tscanner.gettoken: could not get token  from tokenzier');}
        ErrorMessage := Tokenizers[CT]^.ErrorMessage;
        exit;
      end;
      { check if token is a directive }
      if (T^.MyType = TOK_DIRECTIVE) then
      begin
        if (not IdentifyDirective(T^.Data, DT, P)) then
        begin
          { writeln(GetStreamInfo + 'tscanner.gettoken:
            could not identify directive');}
          Dispose(T, Done);
          Continue;
        end;
        case DT of
          DT_DEFINE:
            AddDirective(P);
          DT_ELSE:
            if (DirectiveLevel > 0) then
            begin
              Dec(DirectiveLevel);
              if (not SkipUntilElseOrEndif(FoundElse))
              then exit;
              if FoundElse
              then ;
            end
            else
            begin
              ErrorMessage := GetStreamInfo + ': unexpected $else directive.';
              exit;
            end;
          DT_ENDIF:
          begin
            Print(5, 'ENDIF encountered');
            if (DirectiveLevel > 0) then
            begin
              Dec(DirectiveLevel);
            end
            else
            begin
              ErrorMessage := GetStreamInfo + ': unexpected $endif directive.';
              exit;
            end;
          end;
          DT_IFDEF:
          begin
            Print(6, 'IFDEF encountered (' + P + '):'); 
            if IsDirectiveDefined(P) then
            begin
              PrintLn(6, ' defined.');
              Inc(DirectiveLevel);
            end
            else
			begin
              PrintLn(6, ' not defined.');
              if (not SkipUntilElseOrEndif(FoundElse))
              then exit;
              if FoundElse
              then Inc(DirectiveLevel);
		    end;
          end;
          DT_IFNDEF:
          if (not IsDirectiveDefined(P))
          then Inc(DirectiveLevel)
          else
          begin
            if (not SkipUntilElseOrEndif(FoundElse))
            then exit;
          end;
          DT_IFOPT:
          begin
            if (not IsSwitchDefined(P)) then
			begin
              if (not SkipUntilElseOrEndif(FoundElse))
              then exit;
              if FoundElse
              then Inc(DirectiveLevel);
		    end
            else Inc(DirectiveLevel);
          end;
          DT_INCLUDE_FILE:
                if (not OpenInclude(P)) then
                begin
                  PrintLn(2, GetStreamInfo +
                    ': Error, could not open include file "' + P + '"');
                  exit;
                end;
          DT_UNDEF:
			  	DeleteDirective(P);
        end;
	  end;
      if (T^.MyType = TOK_DIRECTIVE)
      then Dispose(T, Done)
      else Finished := true;
    end
    else
    begin
      PrintLn(5, 'Closing stream "' + Tokenizers[CT]^.GetStreamInfo + '"');
      Dispose(Tokenizers[CT], Done);
      Tokenizers[CT] := nil;
      Dec(CT);
    end;
  until Finished;
  GetToken := true;
end;

function TScanner.IsDirectiveDefined(N: String): Boolean;
begin
  StringToUpper(N, N);
  IsDirectiveDefined := Assigned(Directives) and
    (Directives^.IndexOf(@N) <> -1);
end;

function TScanner.IsSwitchDefined(N: String): Boolean;
var
  B1: Boolean;
  B2: Boolean;
  L: TLowerCaseLetter;
begin
  { we expect a length 2 string like 'I+' or 'A-', first character a letter,
    second character plus or minus }
  if ((N[1] >= 'A') and (N[1] <= 'Z'))
  then N[1] := Chr(Ord(N[1]) + 32);
  if (Length(N) < 2) or
     ((N[1] < 'a') and (N[1] > 'z')) or
     ((N[2] <> '-') and (N[2] <> '+')) then
  begin
  	PrintLn(2, GetStreamInfo + ': Warning - invalid $ifopt parameters.');
    IsSwitchDefined := false;
  end
  else
  begin
    { look up switch from current table }
    L := TLowerCaseLetter(Ord(N[1]) - Ord('a'));
    B1 := SwitchOptions[L];
    { get status from parameter }
    B2 := (N[2] = '+');
    IsSwitchDefined := (B1 = B2);
  end;
end;

function TScanner.OpenInclude(N: String): Boolean;
var
  I: Integer;
  Name: String;
  NumAttempts: Integer;
  P: PString;
  S: PFileInputStream;
begin
  OpenInclude := false;
  { check if maximum number of tokenizers has been reached }
  if (CT = MAX_TOKENIZERS - 1) then
  begin
    ErrorMessage := GetStreamInfo + ': maximum number of tokenizers reached.';
    exit;
  end;
  { determine how many names we can check; number is 1 + IncludeFilePaths^.Count }
  NumAttempts := 1;
  if Assigned(IncludeFilePaths) and (IncludeFilePaths^.Count > 0)
  then Inc(NumAttempts, IncludeFilePaths^.Count);
  { loop until we have checked all names or one attempt was successful }
  for I := 0 to NumAttempts - 1 do
  begin
  	if (I = 0)
  	then Name := N
  	else
  	begin
  	  P := IncludeFilePaths^.At(I - 1);
  	  if Assigned(P)
  	  then Name := P^ + N
  	  else Continue; { next loop iteration }
  	end;
    PrintLn(5, 'Trying to open include file "' + Name + '"...');
    S := New(PFileInputStream, Init(Name));
    if Assigned(S)
    then Break;
  end;
  { if we still don't have a valid open stream we failed }
  if (not Assigned(S)) then
  begin
    ErrorMessage := GetStreamInfo + ': could not open include file ' + N;
    exit;
  end;
  { create new tokenizer with stream }
  Tokenizers[CT + 1] := New(PTokenizer, Init(S));
  if (not Assigned(Tokenizers[CT + 1])) then
  begin
    ErrorMessage := GetStreamInfo + ': could not create tokenizer.';
    exit;
  end;
  Inc(CT);
  OpenInclude := true;
end;

function TScanner.PeekToken(var T: PToken): Boolean;
begin
  if IsTokenBuffered
  then
    begin
      T := BufferedToken;
      PeekToken := true;
    end
  else
    begin
      if GetToken(T)
      then
        begin
          IsTokenBuffered := true;
          BufferedToken := T;
          PeekToken := true;
        end
      else PeekToken := false;
    end;
end;

function TScanner.SkipUntilElseOrEndif(var FoundElse: Boolean): Boolean;
var
  DT: TDirectiveType;
  Level: Integer;
  P: String;
  T: PToken;
  TT: TTokenType;
begin
  SkipUntilElseOrEndif := false;
  Level := 1;
  repeat
    if (not Tokenizers[CT]^.GetToken(T)) then
    begin
      WriteLn('error skipuntilelse gettoken');
      exit;
    end;
    {writeln('token ', T^.Data^.GetString);}
    if (T^.MyType = TOK_DIRECTIVE) then
    begin
      if IdentifyDirective(T^.Data, DT, P) then
      begin
        PrintLn(6, 'SkipUntilElseOrFound: encountered directive ' + DIRECTIVENAMES[Ord(DT)]);
        case DT of
          DT_IFDEF,
          DT_IFNDEF,
          DT_IFOPT: Inc(Level);
          DT_ELSE,
          DT_ENDIF: Dec(Level);
        end;
      end;
    end;
    TT := T^.MyType;
    Dispose(T, Done);
  until (Level = 0) and (TT = TOK_DIRECTIVE) and ((DT = DT_ELSE) or (DT = DT_ENDIF));
  FoundElse := (DT = DT_ELSE);
  if FoundElse
  then P := 'ELSE'
  else P := 'ENDIF';
  PrintLn(6, 'Skipped code, last token ' + P);
  SkipUntilElseOrEndif := true;
end;

procedure TScanner.UnGetToken(T: PToken);
begin
  if IsTokenBuffered then
  begin
    PrintLn(1, GetStreamInfo + ': FATAL ERROR - CANNOT UNGET MORE THAN ONE TOKEN.');
    Halt(1);
  end;
  IsTokenBuffered := true;
  BufferedToken := T;
end;

procedure TScanner.UpdateSwitchOptions(O: String);
begin
  { assume dollar sign has been removed }
end;

begin
end.
