{
@abstract(provides LaTeX doc generator object)
@author(Michael van Canneyt (michael@tfdec1.fys.kuleuven.ac.be))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(9 Sep 1998)
@lastmod(10 Apr 2000)

Implements an object to generate TeX documentation, overriding many of
@link(TDocGenerator)'s virtual methods.
TeX is a free text setting package that produces output well-suited for
printing.
}

unit tex;

interface

uses
  appinfo, arrays, filestre, gendoc, items, objects, parsing, streams, texts,
  time;

const
  { name of the TeX output file (there is just a single file for all the
    documentation) }
  TEX_OUTPUT_FILENAME = 'docs';

type
  { pointer to @link(TTexDocGenerator) }
  PTexDocGenerator = ^TTexDocGenerator;
  { @abstract(generates TeX documentation)
    Extends @link(TDocGenerator) and overwrites many of its methods to
    generate output in TeX format.
    This type of output is well suited to be printed. }
  TTexDocGenerator = object(TDocGenerator)
    { number of cells we've already written in current table row }
    CellCounter: LongInt;
    { default is false; if true pasdoc will not write document type at
      the beginning or end-of-document information at the end so that
      one can include the .tex output file in any other tex document }
    NoTexHeadersAndFooters: Boolean;
    { number of cells (= columns) per table row }
    NumCells: LongInt;
    { Implements @link(TDocGenerator.ConvertChar).
      Returns a string that can be used to output character C in TeX.
      Several special characters must be escaped, e.g. the backslash. }
    function ConvertChar(C: Char): String; virtual;
    { Returns a link to an anchor within a document. }
    function CreateLink(Item: PItem): String; virtual;
    function CreateReferencedLink(ItemName, Link: String): String; virtual;
    { Returns (La)TeX file extension '.tex'. }
    function GetFileExtension: String; virtual;
    { Writes an anchor, i.e., the name of the label by which item I will be
      referenced in other item's comments.
      Will be constructed from @link(TItem.AnchorNumber) and
      @link(AnchorToString). }
    procedure WriteAnchor(I: PItem);
    { Writes all information of a single class, interface or object (CIO)
      to output, starting at heading level HL. }
    procedure WriteCIO(HL: Byte; var CIO: TCIO); virtual;
    { Writes information on all classes, interfaces and objects from C to
      output, starting at heading level HL. }
    procedure WriteCIOs(HL: Byte; C: PItemCollection); virtual;
    { Writes a comment to output file. }
    procedure WriteComment(const S: String); virtual;
    { Writes all documentation files - number will vary between output
      formats. }
    procedure WriteDocumentation; virtual;
    { original comment }
    procedure WriteFields(HL: Byte; C: PItemCollection);
    { Writes functions and procedures / methods from FuncsProcs to the docs,
      at heading level HL. If methods is true, it's a set of methods,
      otherwise a set of functions and procedures. }
    procedure WriteFuncsProcs(HL: Byte; Methods: Boolean; FuncsProcs:
      PMethodCollection); virtual;
    { Writes a heading S at heading level HL.
      HL should be from 1 to 4.
      The smaller the number, the larger will be the resulting heading. }
    procedure WriteHeading(HL: Byte; const S: String); virtual;
    { Writes descriptions for all items in the given collection (e.g.
      variables, constants etc.) at heading level HL with given heading. }
    procedure WriteItems(HL: Byte; Heading: String; I: PItemCollection); virtual;
    procedure WriteMethodOverview(P: PMethodCollection); virtual;
    procedure WriteMethodDeclaration(const S: String); virtual;
    procedure WriteText(T: PText); virtual;
    procedure WriteProperties(HL: Byte; P: PPropertyCollection); virtual;
    { Writes necessary introductory statements for a valid LaTeX document,
      includes several packages (e.g. for tables). }
    procedure WriteStartOfDocument;
    { Writes all documentation for one unit. }
    procedure WriteUnit(HL: Byte; U: PUnit); virtual;
    procedure WriteUnitDescription(HL: Byte; D: PText); virtual;
  private
    {  }
    procedure WriteEndOfAnchor(const Name: String);
    {  }
    procedure WriteEndOfDocument;
    procedure WriteEndOfLink(const Name: String); virtual;
    procedure WriteEndOfTable2Columns;
    procedure WriteEndOfTableCell;
    procedure WriteEndOfTableRow;
    procedure WriteStartOfLink(const Name: String);
    procedure WriteStartOfTable1Column(T: String);
    procedure WriteStartOfTable2Columns(T1, T2: String);
    procedure WriteStartOfTable3Columns(T1, T2, T3: String);
    procedure WriteStartOfTableRow;
  end;

implementation

uses
  msg;

function TTexDocGenerator.ConvertChar(C: Char): String;
begin
  if (C in ['&','^','_','$','%','#','{','}'])
  then ConvertChar := '\' + C
  else
  if (C in ['<', '>'])
  then ConvertChar := '$' + C + '$'
  else
    begin
      if (C = '\')
      then ConvertChar := '\verb|\|'
      else ConvertChar := C;
    end;
end;

function TTexDocGenerator.CreateLink(Item: PItem): String;
begin
  CreateLink := AnchorToString(Item^.AnchorNumber);
end;

function TTexDocGenerator.CreateReferencedLink(ItemName, Link: String): String;
begin
  CreateReferencedLink := '{\ttfamily ' + ConvertString(ItemName) +
    '} (p. \pageref{' + Link + '})';
end;

function TTexDocGenerator.GetFileExtension: String;
begin
  GetFileExtension := '.tex';
end;

procedure TTexDocGenerator.WriteCIO(HL: Byte; var CIO: TCIO);
var
  S: String;
begin
  case CIO.MyType of
    CIO_CLASS: S := GetString(TEXT_CLASS);
    CIO_INTERFACE: S := GetString(TEXT_INTERFACE);
    CIO_OBJECT: S := GetString(TEXT_OBJECT);
    else
      begin
        PrintLn(1, 'Fatal error in TTexDocGenerator.WriteCIO: unknown CIO type.');
        Halt(1);
      end;
  end;
  S := S + ' ' + CIO.Name;
  WriteAnchor(@CIO);
  WriteDescription(HL + 1, S, CIO);
  WriteFields(HL + 1, CIO.Fields);
  WriteFuncsProcs(HL + 1, true, CIO.Methods);
  WriteProperties(HL + 1, CIO.Properties);
end;

procedure TTexDocGenerator.WriteCIOs(HL: Byte; C: PItemCollection);
var
  I: LongInt;
  Item: PCIO;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  PrintLn(4, 'Writing docs for classes/interfaces/objects');
  for I := 0 to C^.Count - 1 do
  begin
    Item := C^.At(I);
    { note that we do _not_ increase the heading level
      we might otherwise run out of heading levels for TeX etc. }
    if Assigned(Item)
    then WriteCIO(HL, Item^);
  end;
end;

procedure TTexDocGenerator.WriteComment(const S: String);
begin
  Stream^.WriteChar('%');
  Stream^.WriteLine (S)  
end;

procedure TTexDocGenerator.WriteDocumentation;
begin
  if (not CreateStream(TEX_OUTPUT_FILENAME)) then
  begin
    PrintLn(1, 'Error - could not create output stream.');
    exit;
  end;
  WriteComment(APP_FIRST_LINE);
  WriteComment(GetString(TEXT_WARNING_AUTOMATIC));
  WriteStartOfDocument;
  WriteUnits(1);
  WriteEndOfDocument;
  CloseStream;
end;

procedure TTexDocGenerator.WriteEndOfDocument;
begin
  if NoTexHeadersAndFooters
  then exit;
  PrintLn(4, 'Writing end of TeX document.');
  if (not Assigned(Stream)) then
  begin
    PrintLn(2, 'Error TTexDocGenerator.WriteEndOfDocument: No stream available.');
    exit;
  end;
  Stream^.WriteLine('\end{document}');
end;

procedure TTexDocGenerator.WriteEndOfAnchor(const Name: String);
begin
end;

procedure TTexDocGenerator.WriteEndOfLink(const Name: String);
begin
end;

(*
procedure TTexDocGenerator.WriteEndOfTable1Column;
begin
  Stream^.WriteLine('\hline');
  Stream^.WriteLine('\end{tabular}');
end;*)

procedure TTexDocGenerator.WriteEndOfTable2Columns;
begin
        { If you insert again the \hline at the endofrow,
          then remove this one }
  Stream^.WriteLine('\hline');
  Stream^.WriteLine('\end{tabularx}');
end;

(*procedure TTexDocGenerator.WriteEndOfTable3Columns;
begin
        { If you insert again the \hline at the endofrow,
          then remove this one }
  Stream^.WriteLine('\hline');
  Stream^.WriteLine('\end{tabularx}');
end;*)

procedure TTexDocGenerator.WriteEndOfTableCell;
begin
  Inc(CellCounter);
  if (CellCounter < NumCells)
  then Stream^.WriteString(' & ');
end;

procedure TTexDocGenerator.WriteEndOfTableRow;
begin
  Stream^.WriteLine(' \\');
{  Stream^.WriteLine('\hline');}
end;

procedure TTexDocGenerator.WriteFields(HL: Byte; C: PItemCollection);
begin
  WriteItems(HL, GetString(TEXT_FIELDS), C);
end;

procedure TTexDocGenerator.WriteFuncsProcs(HL: Byte; Methods: Boolean;
  FuncsProcs: PMethodCollection);
var
  I: LongInt;
  M: PMethod;
begin
  if (not Assigned(FuncsProcs)) or (FuncsProcs^.Count < 1)
  then exit;
  PrintLn(4, 'Writing funcs/procs/methods.');
  if Methods
  then WriteHeading(HL, GetString(TEXT_METHODS))
  else WriteHeading(HL, GetString(TEXT_FUNCTIONS_AND_PROCEDURES));
  for I := 0 to FuncsProcs^.Count - 1 do
  begin
    M := FuncsProcs^.At(I);
    { write a label that can be referenced elsewhere in the output file }
    WriteAnchor(M);
    { write method declaration in courier font, then line feed }
    Stream^.WriteString('{\ttfamily ');
    WriteString(M^.FullDeclaration);
    Stream^.WriteLine('}\\');
    { write description of method }
    WriteText(M^.GetDescription);
    { write line feed in output so that next declaration will be at
      the beginning of a new line, with one line distance }
    Stream^.WriteLine('\\');
    Stream^.WriteLine('\\');
  end;
end;

procedure TTexDocGenerator.WriteHeading(HL: Byte; const S: String);
begin
  if (not Assigned(Stream)) then
  begin
    PrintLn(2, 'Error TTexDocGenerator.WriteHeading: No stream available.');
    exit;
  end;
  case HL of
    1: Stream^.WriteString('\chapter*{');
    2: Stream^.WriteString('\section*{');
    3: Stream^.WriteString('\subsection*{');
    4: Stream^.WriteString('\subsubsection*{');
    else Stream^.WriteChar('{');
  end;
  WriteString(S);
  Stream^.WriteLine('}');
end;

procedure TTexDocGenerator.WriteItems(HL: Byte; Heading: String;
  I: PItemCollection);
var
  J: LongInt;
  Item: PItem;
begin
  if (not Assigned(I)) or (I^.Count < 1)
  then exit;
  PrintLn(4, 'Writing items (' + Heading + ')');
  WriteHeading(HL, Heading);
  for J := 0 to I^.Count - 1 do
  begin
    Item := I^.At(J);
    WriteAnchor(Item);
    Stream^.WriteString('\noindent ');
    Stream^.WriteString('{\ttfamily ');
    WriteString(Item^.Name);
    Stream^.WriteString('}');
    Stream^.WriteLine('\\');
    WriteText(Item^.GetDescription);
    Stream^.WriteLine('\\');
  end;
end;

Procedure TTexDocGenerator.WriteMethodOverview(P: PMethodCollection);
Var
  I: LongInt;
  Item: PMethod;
begin
  if (not Assigned(P)) or (P^.Count < 1)
  then exit;
  WriteStartOfTable2Columns('Name','Page');
  for I := 0 to P^.Count - 1 do
  begin
    Item := P^.At(I);
    WriteStartOfTableRow;
    WriteMethodDeclaration(Item^.FullDeclaration);
    WriteEndOfTableCell;
    WriteStartOfLink(Item^.FullLink);
    WriteEndOfLink(Item^.FullLink);
    WriteEndOfTableCell;
    WriteEndOfTableRow;
  end;
  WriteEndOfTable2Columns;
end;

procedure TTexDocGenerator.WriteMethodDeclaration(const S: String);
begin
        { the \ttfamily jumps to teletype font.}
  Stream^.WriteString('{\ttfamily ');
  WriteString(S);
  Stream^.WriteLine('} \par');
end;

procedure TTexDocGenerator.WriteProperties(HL: Byte; P: PPropertyCollection);
begin
end;

(*
Procedure TTexDocGenerator.WritePropertyOverview(P: PPropertyCollection);
Var
  I: LongInt;
  Item: PProperty;
begin
  if (not Assigned(P)) or (P^.Count < 1)
  then exit;
  WriteStartOfTable2Columns('Name','Page');
  for I := 0 to P^.Count - 1 do
  begin
    Item := P^.At(I);
    WriteStartOfTableRow;
    WritePropertyDeclaration(Item);
    WriteEndOfTableCell;
    WriteStartOfLink(Item^.FullLink);
    WriteEndOfLink(Item^.FullLink);
    WriteEndOfTableCell;
    WriteEndOfTableRow;
  end;
  WriteEndOfTable2Columns;
end;

procedure TTexDocGenerator.WritePropertyDeclaration(P : PProperty);
begin
  Stream^.WriteString('{\ttfamily ');
  WriteString(p^.GetDeclaration);
  Stream^.WriteLine('} \par');
end;
*)

procedure TTexDocGenerator.WriteAnchor(I: PItem);
begin
  Stream^.WriteString('\label{' + AnchorToString(I^.AnchorNumber) + '}');
end;

procedure TTexDocGenerator.WriteStartOfDocument;
begin
  if NoTexHeadersAndFooters
  then exit;
  PrintLn(4, 'Writing start of TeX document.');
  if (not Assigned(Stream)) then
  begin
    PrintLn(2, 'Error in TTexDocGenerator.WriteStartOfDocument: No stream available.');
    exit;
  end;
  Stream^.WriteLine('\documentclass{report}');
  Stream^.WriteLine('\usepackage{tabularx}');
  Stream^.WriteLine('\begin{document}');
end;

procedure TTexDocGenerator.WriteStartOfLink(const Name: String);
begin
  Stream^.WriteString('(\pageref{' + Name + '})');
end;

procedure TTexDocGenerator.WriteStartOfTable1Column(T: String);
begin
  Stream^.WriteLine('\begin{tabular}{|l|}');
  Stream^.WriteLine('\hline');
  NumCells := 1;
end;

procedure TTexDocGenerator.WriteStartOfTable2Columns(T1, T2: String);
begin
  Stream^.WriteLine('\begin{tabularx}{\linewidth}{|lX|}');
  Stream^.WriteLine('\hline');
  WriteString(T1);
  Stream^.WriteString(' & ');
  WriteString(T2);
  Stream^.WriteLine(' \\');
  Stream^.WriteLine('\hline');
  NumCells := 2;
end;

procedure TTexDocGenerator.WriteStartOfTable3Columns(T1, T2, T3: String);
begin
  Stream^.WriteLine('\begin{tabularx}{\linewidth}{|llX|}');
  Stream^.WriteLine('\hline');
  WriteString(T1);
  Stream^.WriteString(' & ');
  WriteString(T2);
  Stream^.WriteString(' & ');
  WriteString(T3);
  Stream^.WriteLine(' \\');
  Stream^.WriteLine('\hline');
  NumCells := 3;
end;

procedure TTexDocGenerator.WriteStartOfTableRow;
begin
  CellCounter := 0;
end;

procedure TTexDocGenerator.WriteText(T: PText);
var
  C: Char;
  I: LongInt;
  L: LongInt;
begin
  if (T = nil) or (T^.Content < 1)
  then exit;
  L := 0;
  for I := 0 to T^.Content - 1 do
  begin
  	C := T^.Data^[I];
  	if (C = #9) or (C = #10) or (C = #13)
  	then C := ' ';
    Stream^.WriteChar(C);
    Inc(L);
    if (L > 80) and (T^.Data^[I] = ' ')
    then
      begin
      	Stream^.WriteLine('');
      	L := 0;
      end;
  end;
{  Stream^.WriteLine('\\');}
end;

procedure TTexDocGenerator.WriteUnit(HL: Byte; U: PUnit);
begin
  if (not Assigned(U))
  then exit;
  PrintLn(3, 'Writing docs for unit ' + U^.Name);
{  if (Assigned(U^.DetailedDescription))
  then D := U^.DetailedDescription
  else D := U^.Description;
  WriteUnitDescription(HL, D);}
  WriteDescription(HL, GetString(TEXT_UNIT) + ' ' + U^.Name, U^);
  WriteConstants(HL + 1, U^.Constants);
  WriteVariables(HL + 1, U^.Variables);
  WriteTypes(HL + 1, U^.Types);
  WriteFuncsProcs(HL + 1, false, U^.FuncsProcs);
  WriteCIOs(HL + 1, U^.CIO);
end;

procedure TTexDocGenerator.WriteUnitDescription(HL: Byte; D: PText);
var
  I: LongInt;
begin
  if (not Assigned(D)) or (D^.Content < 1) or (not Assigned(D^.Data))
  then exit;
  PrintLn(4, 'Writing unit description.');
  WriteHeading(HL, GetString(TEXT_DESCRIPTION));
  for I := 0 to D^.Content - 1 do
  begin
    Stream^.WriteChar(D^.Data^[I]);
  end;
end;

begin
end.
