{
@abstract(provides dynamic-length strings)
@created(22 Feb 1999)
@lastmod(8 Feb 2000)
@author(Marco Schmidt (marcoschmidt@geocities.com))
}
unit texts;

interface

uses
  objects;

const
  { set of characters, including all letters and the underscore }
  IdentifierStart = ['A'..'Z', 'a'..'z', '_'];
  { set of characters, including all characters from @link(IdentifierStart)
    plus the ten decimal digits }
  IdentifierOther = ['A'..'Z', 'a'..'z', '_', '0'..'9', '.'];
  { set of characters, including tab, space, line feed and carriage return
    (decimal values 9, 10, 13 and 32) }
  Whitespace = [#9, #10, #13, #32];
  { set of characters, including all lowercase and uppercase letters }
  Letters = ['A'..'Z', 'a'..'z'];

type
  { array of characters, which will be overindexed }
  TCharArray = Array[0..99] of Char;
  { pointer to a @link(TCharArray) }
  PCharArray = ^TCharArray;
  { pointer to @link(TText) }
  PText = ^TText;
  { @abstract(simple text buffer, capable of increasing its size dynamically)
    An object to store and manipulate an array of characters, which can
    change its size dynamically.
    Can handle pasdoc tags - a tag is considered a part of a comment which
    starts with the at character at, then has one or more letters.
    The letters make up the name of the tag, then follows an opening
    bracket, the tag content and a closing bracket. }
  TText = object(TObject)
    { number of characters actually stored in @link(TText.Data) }
    Content: LongInt;
    { character buffer }
    Data: PCharArray;
    { current size of character buffer Data }
    Size: LongInt;
    { Initializes all fields to default values. }
    constructor Init;
    { Releases any memory still allocated to @link(Data). }
    destructor Done; virtual;
    { Appends character C to the text buffer @link(Data), which is created
      if not already existing. }
    function AppendChar(C: Char): Boolean;
    { Appends string S to the text buffer @link(Data), which is created
      if not already existing. }
    function AppendString(S: String): Boolean;
    { Removes N characters from field @link(Data) starting at index I. }
    procedure Delete(I, N: LongInt);
    { Deletes all characters from Offs1 to Offs2 (including these two).
      Simply calls @link(Delete) with corresponding parameters.
      Note that you have to take Offs1 and Offs3 from FindTag when
      calling this procedure, not Offs1 and Offs2. }
    procedure DeleteTag(Offs1, Offs2: LongInt);
    { }
    procedure ExtractTag(Offs1, Offs2, Offs3: LongInt; var S: String);
    { Attempts to find tag with given Name, starting at offset Offs1.
      If successful, true is returned and position of opening and closing
      parenthesis in Offs2 and Offs3, false is returned otherwise. }
    function FindTag(Name: String; var Offs1, Offs2, Offs3: LongInt): Boolean;
    { }
    function FindTagParameters(var Offs1, Offs2: LongInt): Boolean;
    { Attempts to return content as string.
      Data may be missing, as strings can only hold up to 255 characters. }
    function GetString: String;
    { returns number of free characters in buffer (Size - Content) }
    function GetNumFreeChars: LongInt;
    {  }
    procedure GetTag(Remove: Boolean; Offs1, Offs2, Offs3: LongInt; var S: String);
    { Gets a tag name, starting at position Offset. If Data^[Offset] is
      not a letter, the returned tag name will be empty! }
    function GetTagName(var Offset: LongInt): String;
    { }
    procedure Print;
    { Attempts to increase size of character buffer to NewSize and
      returns success or failure as a boolean value. }
    function Reallocate(NewSize: LongInt): Boolean;
    { replaces all whitespace with blanks, then removes strings of
      more than one blank and replaces them with a single blank }
    procedure RemoveObsoleteWhitespace;
  end;

implementation

uses
  arrays;

{ TText }

constructor TText.Init;
begin
  inherited Init;
  Content := 0;
  Data := nil;
  Size := 0;
end;

destructor TText.Done;
begin
  if Assigned(Data)
  then FreeMem(Data, Size);
  inherited Done;
end;

function TText.AppendChar(C: Char): Boolean;
const
  STEP = 128;
begin
  if Assigned(Data)
  then
    begin
      if (Content < Size)
      then
        begin
          Data^[Content] := C;
          Inc(Content);
          AppendChar := true;
        end
      else
        begin
          if Reallocate(Size + STEP)
          then AppendChar := AppendChar(C)
          else AppendChar := false;
        end;
    end
  else
    begin
      if Reallocate(Size + STEP)
      then AppendChar := AppendChar(C)
      else AppendChar := false;
    end;
end;

function TText.AppendString(S: String): Boolean;
var
  I: LongInt;
begin
  AppendString := true;
  if (Length(S) < 1)
  then exit;
  for I := 1 to Length(S) do
    if (not AppendChar(S[I]))
    then
      begin
        AppendString := false;
        exit;
      end;
end;

procedure TText.Delete(I, N: LongInt);
begin
  if (I < 0) or (I >= Content)
  then exit;
  if (I + N > Content)
  then N := Content - I;
  System.Move(Data^[I + N], Data^[I], Content - (I + N));
  Dec(Content, N);
end;

procedure TText.DeleteTag(Offs1, Offs2: LongInt);
begin
  Self.Delete(Offs1, Offs2 - Offs1 + 1);
end;

procedure TText.ExtractTag(Offs1, Offs2, Offs3: LongInt; var S: String);
begin
  GetTag(true, Offs1, Offs2, Offs3, S);
end;

{
INPUT
set Offs1 to the index where you want to start the search
OUTPUT
if result is true:
  @abstract(this is the abstract)
  ^    ^          ^
  offs1  offs2        offs3
}
function TText.FindTag(Name: String; var Offs1, Offs2, Offs3: LongInt): Boolean;
var
  I: LongInt;
  J: LongInt;
  S: String;
  U: String;
begin
  FindTag := false;
  if (not Assigned(Data)) or (Offs1 < 0) or (Offs1 >= Content) or
     (Length(Name) < 1)
  then exit;
  Arrays.StringToUpper(Name, Name);
  I := Offs1;
  repeat
    if (Data^[I] = '@')
    then
      begin
        if (I + 1 = Content)
        then exit;
        if (Data^[I + 1] = '@')
        then
          begin
            Inc(I, 2);
            Continue;
          end;
        Offs1 := I;
        J := I + 1;
        S := GetTagName(J);
        StringToUpper(S, U);
        Offs2 := J;
        if (not FindTagParameters(Offs2, Offs3))
        then exit;
        if (Name = U)
        then
          begin
            FindTag := true;
            exit;
          end;
        I := Offs3 + 1;
      end
    else Inc(I);
  until (I = Content);
end;

function TText.FindTagParameters(var Offs1, Offs2: LongInt): Boolean;
var
  Counter: LongInt;
  I: LongInt;
begin
  FindTagParameters := false;
  I := Offs1;
  if (I < 0) or (I >= Content)
  then exit;
  while (I < Content) and (Data^[I] <> '(') do
    Inc(I);
  if (I = Content) { no ( found }
  then exit;
  Offs1 := I;
  Inc(I);
  Counter := 1;
  repeat
    case Data^[I] of
      '(': Inc(Counter);
      ')': Dec(Counter);
    end;
    Inc(I);
  until (I = Content) or (Counter = 0);
  if (Counter = 0)
  then
    begin
      Offs2 := I - 1;
      FindTagParameters := true;
    end;
end;

function TText.GetNumFreeChars: LongInt;
begin
  GetNumFreeChars := (Size - Content);
end;

function TText.GetString: String;
var
  L: Integer;
  S: String[255];
begin
  if (Content > 0)
  then
    begin
      if (Content <= 255)
      then L := Content
      else L := 255;
      System.Move(Data^, S[1], L);
      SetLength(S, L);
      GetString := S;
    end
  else GetString := '';
end;

procedure TText.GetTag(Remove: Boolean; Offs1, Offs2, Offs3: LongInt; var S: String);
var
  L: LongInt;
begin
  L := (Offs3 - Offs2 - 1);
  if (L < 0)
  then L := 0;
  if (L > 255)
  then L := 255;
  System.Move(Data^[Offs2 + 1], S[1], L);
  SetLength(S, L);
  if Remove
  then Self.Delete(Offs1, Offs3 - Offs1 + 1);
end;

function TText.GetTagName(var Offset: LongInt): String;
var
  S: String;
begin
  S := '';
  while (Offset < Content) and
    (((Data^[Offset] >= 'A') and (Data^[Offset] <= 'Z')) or
     ((Data^[Offset] >= 'a') and (Data^[Offset] <= 'z'))) do
  begin
    S := S + Upcase(Data^[Offset]);
    Inc(Offset);
  end;
  GetTagName := S;
end;

procedure TText.Print;
var
	I: LongInt;
begin
	I := 0;
    Write('"');
    while (I < Content) do
    begin
    	Write(Data^[I]);
    	Inc(I);
    end;
    WriteLn('"');
end;

function TText.Reallocate(NewSize: LongInt): Boolean;
var
  Temp: PCharArray;
begin
  if (NewSize <= Size) then
  begin
    Reallocate := true;
    exit;
  end;
  //WriteLn('before getmem, newsize=', newsize, ', size=', size);
  GetMem(Temp, NewSize);
  //WriteLn('after getmem');
  if Assigned(Temp) then
  begin
    System.Move(Data^, Temp^, Content);
    FreeMem(Data, Size);
    Data := Temp;
    Size := NewSize;
    Reallocate := true;
  end
  else Reallocate := false;
end;

procedure TText.RemoveObsoleteWhitespace;
var
  I: LongInt;
  J: LongInt;
begin
  if (not Assigned(Data))
  then exit;
  I := 0;
  while (I < Content - 1) do
  begin
    if (Data^[I] in Whitespace)
    then
      begin
        J := 0;
        while (I + J + 1 < Content) and (Data^[I + J + 1] in Whitespace) do
          Inc(J);
        if (J > 0)
        then Self.Delete(I, J);
        Data^[I] := ' ';
      end;
    Inc(I);
  end;
  if (Content > 0) and (Data^[0] in Whitespace)
  then Self.Delete(0, 1);
  if (Content > 0) and (Data^[Content - 1] in Whitespace)
  then Self.Delete(Content - 1, 1);
end;

{function TText.SkipTag(Offset: LongInt): LongInt;
var
  Counter: LongInt;
begin
  SkipTag := -1;
  if (not Assigned(Data)) or (Offset >= Content - 1) or
     (Data^[Offset] <> '@') or (not (Data^[Offset + 1] in IdentifierStart))
  then exit;
  Inc(Offset);
  while (Offset < Content) and (Data^[Offset] <> '(') do
    Inc(Offset);
  if (Offset = Content)
  then exit;
  Counter := 1;
  repeat
    case Data^[Offset] of
      '(': Inc(Counter);
      ')': Dec(Counter);
    end;
    Inc(Offset);
  until (Counter = 0) or (Offset = Content);
  if (Counter = 0)
  then SkipTag := Offset;
end;}


begin
end.