{
@abstract(provides time manipulation object)
@author(Marco Schmidt (marcoschmidt@geocities.com))
@lastmod(7 Feb 2000)

The TTime class defined herein represents a single point in time,
e.g. 31 Dec 1998 13:00:03 GMT.

TTimeOptions defines how a TTime object will be converted to a
string or initialized from a string.
}
unit time;

interface

{$I platform.inc}

uses
  {$IFDEF PPC_DELPHI}
  SysUtils,
  {$ENDIF}
  numbers;

type
	{ pointer to @link(TTimeOptions) }
    PTimeOptions = ^TTimeOptions;
	{ @abstract(stores information on how a @link(TTime) object is converted
	  to String)

      In this object you specify how a @link(TTime) object is converted to
      string.

      Define the field @link(TimeString) the way you want a point in time to
	  be converted to a string.

      Take the % plus character combinations to be replaced by the actual
	  values in @link(TTime.GetString) - this way you have the absolute
      freedom of what you want to have included.

	  %Y  Full year (1998)
	  %y  Short year (98)

	  %M  Two digit month number (03)
	  %m  Month number (3)
	%N  Long month name (March)
	%n  Short month name (Mar)

	%W	Long day-of-the-week name (Tuesday)
	%w	Short day-of-the-week name (Tue)

	%D  Two digit day number (09)
	%d  Day number (9)

	%H  Two digit hour number(05)
	%h	Hour number (5)

	%L  Two digit minute number(05)
	%l	Minute number (5)

	%S  Two digit second number(02)
	%s	Minute number (2)

	%E  'am' or 'pm'
	%Z	Time zone 'GMT', 'EST', 'PST' etc.
	%%  a literal %
	}
	TTimeOptions = object
		{ 0=Sunday, 1=Monday, ..., 6=Saturday }
		FirstDOW: Integer;
		{ format of a string with date and time }
		TimeString: String;
	 	{ difference in seconds to GMT }
		TimeZone: TSInt32;
		{ Initialize fields to default values }
		constructor Init;
	end;

var
	{ default options to be used in @link(TTime.GetDefaultString);
	  it will be initialized in the unit initialization section;
	  this option variable will create output of the form "Wed 31 Dec 1998
	  13:12:34" }
	CurrentTimeOptions: TTimeOptions;

type
	{ pointer to @link(TTime) }
	PTime = ^TTime;
	{ Stores a single point in time and allows manipulation and conversion
	  to / from string. Note: 2 author tags within object definition!
	  @author(Marco Schmidt <marcoschmidt@geocities.com>)
	  @author(John Doe <doe@anonymous.org>) }
	TTime = object
		{ Initializes fields, by calling @link(SetCurrentDateTime). }
		constructor Init;
		{ Compares the point of time stored in this object to that in DT. }
		function CompareTo(var DT: TTime): TSInt32;
		{ Converts date in D, M and Y to days passed since Jan 1, 1600,
		  which will be stored in NumDays; returns true if date was valid,
		  false otherwise. }
		function ConvertDMYToDays(D, M, Y: TSInt32; var NumDays: TSInt32): Boolean;
		{ Like @link(ConvertDMYToDays), just converts in the opposite
		  direction. }
		function ConvertDaysToDMY(NumDays: TSInt32; var D, M, Y: TSInt32): Boolean;
		{ Decreases point of time stored in this object by N days. }
		procedure DecDays(N: TSInt32);
		{ Decreases point of time stored in this object by N hours. }
		procedure DecHours(N: TSInt32);
		{ Decreases point of time stored in this object by N minutes. }
		procedure DecMinutes(N: TSInt32);
		{ Decreases point of time stored in this object by N seconds. }
		procedure DecSeconds(N: TSInt32);
		{ Returns difference in days between this object and DT. }
		function GetDiffDays(var DT: TTime): TSInt32;
		{ Returns difference in seconds between this object and DT. }
		function GetDiffSeconds(var DT: TTime): TSInt32;
		{ Returns day of the point of time stored in this object. }
		function GetDay: TSInt32;
		{ Returns day of the week of the point of time stored in this
		  object; 0=Sunday, 1=Monday, ... }
		function GetDayOfTheWeek: TSInt32;
		{ Returns string representation of the point of time stored in this
		  object, calling @link(GetString) with @link(CurrentTimeOptions)
		  as parameter. }
		function GetDefaultString: String;
		{ Returns hour of the point of time stored in this object. }
		function GetHour: TSInt32;
		{ Returns milliseconds of the point of time stored in this object
		  (from 0 to 999). }
		function GetMilliSeconds: TSInt32;
		{ Returns minute of the point of time stored in this object (from 0
		  to 59). }
		function GetMinute: TSInt32;
		{ Returns month of the point of time stored in this object (from 1 to
		  12). }
		function GetMonth: TSInt32;
		{ Returns second of the point of time stored in this object (from 0
		  to 59). }
		function GetSecond: TSInt32;
		{ Returns seconds passed since the beginning of the day stored in
		  this object; equal to (GetHour * 3600 + GetMinute * 60 +
		  GetSecond). }
		function GetSeconds: TSInt32;
		{ Returns string representation of the point of time stored in this
		  object, exact format taken from the Options argument. }
		function GetString(var Options: TTimeOptions): String;
		{ Returns week of the year of the point of time stored in this
		  object. The first week of the year must have at least four days
		  in this year, otherwise a date belongs to the last week of the
		  previous year. }
		function GetWeekOfTheYear: TSInt32;
		{ Returns year of the point of time stored in this object (from 1600
		  to ?). }
		function GetYear: TSInt32;
		{ Returns if year stored in this object is a leap year. }
		function IsLeapYear: Boolean;
		{ Sets current system date and time. }
		procedure SetCurrentDateTime;
		{ Sets given date. }
		procedure SetDate(D, M, Y: TSInt32);
		{ Sets given date and time. }
		procedure SetDateTime(D, M, Y, H, Min, S: TSInt32);
		{ sets }
{		procedure SetDateTimeFromString(D: String);}
		{ Sets milliseconds to M. }
		procedure SetMilliSeconds(M: TSInt32);
		{ Sets given time. }
		procedure SetTime(H, M, S: TSInt32);
	private
		Days,
		Day,
		Month,
		Year,
		Hour,
		Minute,
		Second,
		Millis: TSInt32;
	end;

implementation

{$I platform.inc}

{$ifndef PPC_DELPHI}
uses
  DOS; { GetDate, GetTime }
{$endif}

const
	DAYS_PER_400_YEARS = 146097;
	DAYS_PER_4_YEARS = 1461;
	DAYS_PER_WEEK = 7;
	FIRST_YEAR = 1600;
	HOURS_PER_DAY = 24;
	LAST_YEAR = 2999;
	MONTHS_PER_YEAR = 12;
	SECONDS_PER_DAY = 24 * 60 * 60;
	SECONDS_PER_HOUR = 60 * 60;
	SECONDS_PER_MINUTE = 60;
	MonthLengths: Array[0..MONTHS_PER_YEAR - 1] of Byte =
		(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
	LongDOWNames: Array[0..DAYS_PER_WEEK - 1] of String[9] =
		('Sunday', 'Monday', 'Tuesday', 'Wednesday',
		 'Thursday', 'Friday', 'Saturday');
	LongMonthNames: Array[0..MONTHS_PER_YEAR - 1] of String[9] =
		('January', 'February', 'March', 'April', 'May', 'June',
		 'July', 'August', 'September', 'October', 'November', 'December');
	ShortDOWNames: Array[0..DAYS_PER_WEEK - 1] of String[3] =
		('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
	ShortMonthNames: Array[0..MONTHS_PER_YEAR - 1] of String[3] =
		('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
		 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

{ TTime }

constructor TTime.Init;
begin
	SetCurrentDateTime;
end;

function TTime.CompareTo(var DT: TTime): TSInt32;
var
	Pass: TUInt8;
	R: TSInt16;
	V1: TSInt32;
	V2: TSInt32;
begin
	if (Days < DT.Days)
	then 
		begin
			CompareTo := -1;
			exit;
		end
	else
		if (Days > DT.Days)
		then
			begin
				CompareTo := 1;
				exit;
			end;
	{ both have the same day }
	for Pass := 0 to 2 do
	begin
		case Pass of
			0:
			begin
				V1 := GetHour;
				V2 := DT.GetHour;
			end;
			1:
			begin
				V1 := GetMinute;
				V2 := DT.GetMinute;
			end;
			else
			begin
				V1 := GetSecond;
				V2 := DT.GetSecond;
			end;
		end;
		R := CompareUInt16(V1, V2);
		if (R <> 0)
		then
			begin
				CompareTo := R;
				exit;
			end;
	end;
	CompareTo := 0;
end;

function TTime.ConvertDMYToDays(D, M, Y: TSInt32; var NumDays: TSInt32): Boolean;
var
	L: Boolean;
{	R: TSInt32;}
begin
	ConvertDMYToDays := false;
	if (D < 1) or
	   (M < 1) or (M > MONTHS_PER_YEAR) or
	   (Y < FIRST_YEAR) or (Y > LAST_YEAR)
	then exit;
	if (M = 2)
	then
		begin
			L := ((Y mod 4) = 0) and ((Y mod 100) <> 0) and
			   ((Y mod 400) = 0);
			if (L and (D > 29)) or ((not L) and (D > 28))
			then exit;
		end
	else
		begin
			if (D > MonthLengths[M - 1])
			then exit;
		end;
	ConvertDMYToDays := true;
	if (Y = FIRST_YEAR) and (M < 3)
	then
		begin
			if (M = 1)
			then NumDays := D - 1
			else NumDays := D + 30;
			exit;
		end;
	if (M > 2)
	then Dec(M, 3)
	else
		begin
			Inc(M, 9);
			Dec(Y);
		end;
	Dec(Y, FIRST_YEAR);
	NumDays := ((LongInt(Y div 100) * DAYS_PER_400_YEARS) div 4) +
		  ((LongInt(Y mod 100) * DAYS_PER_4_YEARS) div 4) +
		  (((153 * M) + 2) div 5) + D + 59;
end;

function TTime.ConvertDaysToDMY(NumDays: TSInt32; var D, M, Y: TSInt32): Boolean;
begin
	ConvertDaysToDMY := false;
end;

procedure TTime.DecDays(N: TSInt32);
{var
	T1: TSInt32;
	T2: TSInt32;
	T3: TSInt32;}
begin
	if (N >= Days)
	then
		begin
			 Days := Days - N;
			 { we already know the result will be valid,
			   so we ignore the result of the call }
			 ConvertDaysToDMY(Days, Day, Month, Year);
		end
	else
		begin
			{ ERROR: Date out of range }
		end;
end;

procedure TTime.DecHours(N: TSInt32);
var
	I: TSInt32;
begin
	if (N > 0)
	then
		begin
			I := N div HOURS_PER_DAY;
			DecDays(I);
			if (N > Hour)
			then
				begin
					DecDays(1);
					Hour := HOURS_PER_DAY - (N - Hour);
				end
			else Dec(Hour, N);
		end;
end;

procedure TTime.DecMinutes(N: TSInt32);
var
	I: TSInt32;
begin
	if (N > 0)
	then
		begin
			I := N div SECONDS_PER_MINUTE;
			DecHours(I);
		end;
end;

procedure TTime.DecSeconds(N: TSInt32);
var
	I: TSInt32;
begin
	if (N > 0)
	then
		begin
			I := N div 60;
			DecMinutes(I);
			I := N mod 60;
			if (Hour = 0) and (Minute = 0) and (Second < I)
			then
				begin
					DecDays(1);
				end;
		end;
end;

function TTime.GetDiffDays(var DT: TTime): TSInt32;
begin
	GetDiffDays := Days - DT.Days;
end;

function TTime.GetDiffSeconds(var DT: TTime): TSInt32;
var
	S1: TSInt32;
	S2: TSInt32;
begin
	S1 := GetSeconds;
	S2 := DT.GetSeconds;
	GetDiffSeconds := Abs(S2 - S1);
end;

function TTime.GetDay: TSInt32;
begin
	GetDay := Day;
end;

function TTime.GetDayOfTheWeek: TSInt32;
begin
	 GetDayOfTheWeek := (6 + Days) mod 7;
end;

function TTime.GetDefaultString: String;
begin
	GetDefaultString := GetString(CurrentTimeOptions);
end;

function TTime.GetHour: TSInt32;
begin
	GetHour := Hour;
end;

function TTime.GetMilliSeconds: TSInt32;
begin
	GetMilliSeconds := Millis;
end;

function TTime.GetMinute: TSInt32;
begin
	GetMinute := Minute;
end;

function TTime.GetMonth: TSInt32;
begin
	GetMonth := Month;
end;

function TTime.GetSecond: TSInt32;
begin
	GetSecond := Second;
end;

function TTime.GetSeconds: TSInt32;
begin
	GetSeconds := Hour * 3600 + Minute * 60 + Second; { 0 .. 86399 }
end;

function TTime.GetString(var Options: TTimeOptions): String;
var
	I: LongInt;
	L: LongInt;
	S: String;
	T: String;
begin
	I := 1;
	L := Length(CurrentTimeOptions.TimeString);
	S := '';
	while (I <= L) do
	begin
		if (CurrentTimeOptions.TimeString[I] = '%')
		then
			begin
				if (I = L)
				then Continue;
				Inc(I);
				case CurrentTimeOptions.TimeString[I] of
					'D': T := IntToStringLZ(2, Day);
					'H': T := IntToStringLZ(2, Hour);
					'L': T := IntToStringLZ(2, Minute);
					'M': T := IntToStringLZ(2, Month);
					'N': T := LongMonthNames[Month - 1];
					'S': T := IntToStringLZ(2, Second);
					'W': T := LongDOWNames[GetDayOfTheWeek];
					'Y': T := IntToString(Year);
					'd': T := IntToString(Day);
					'h': T := IntToString(Hour);
					'l': T := IntToString(Minute);
					'm': T := IntToString(Month);
					'n': T := ShortMonthNames[Month - 1];
					's': T := IntToString(Second);
					'w': T := ShortDOWNames[GetDayOfTheWeek];
					'y': T := IntToString(Year mod 100);
					'%': T := '%';
					else T := 'ERROR IN OPTIONS.DATETIMESTRING - UNKNOWN ESCAPE CHARACTER: ' +
						CurrentTimeOptions.TimeString[I];
				end;
				S := S + T;
			end
		else S := S + CurrentTimeOptions.TimeString[I];
		Inc(I);
	end;
	GetString := S;
{	%E  'am' or 'pm'
	%Z	Time zone 'GMT', 'EST', 'PST' etc.}
end;

function TTime.GetWeekOfTheYear: TSInt32;
begin
	GetWeekOfTheYear := -1;
end;

function TTime.GetYear: TSInt32;
begin
	GetYear := Year;
end;

function TTime.IsLeapYear: Boolean;
begin
	IsLeapYear :=
		((Year mod 4) = 0) and
		((Year mod 100) <> 0) and
		((Year mod 400) = 0);

{		  LeapYear := ((Year mod 4) = 0) and
			   (not (((Year mod 100)  = 0) and
			   ((Year mod 400) <> 0)));}

end;

{ TODO: test Delphi part with other versions than 5.0 }
procedure TTime.SetCurrentDateTime;
var
  W: Array[0..3] of Word;
{$IFDEF PPC_DELPHI}
  DT: TDateTime;
{$ENDIF}
begin
{$ifdef PPC_DELPHI}
    DT := SysUtils.Date;
    DecodeDate(DT, W[0], W[1], W[2]);
	SetDate(W[2], W[1], W[0]);
    DT := SysUtils.Time;
    DecodeTime(DT, W[0], W[1], W[2], W[3]);
	SetTime(W[0], W[1], W[2]);
	SetMilliSeconds(W[3]);
{$else}
	DOS.GetDate(W[0], W[1], W[2], W[3]);
	SetDate(W[2], W[1], W[0]);
	DOS.GetTime(W[0], W[1], W[2], W[3]);
	SetTime(W[0], W[1], W[2]);
	SetMilliSeconds(W[3] * 10);
{$endif}
end;

procedure TTime.SetDate(D, M, Y: TSInt32);
var
	TempDays: TSInt32;
begin
	if ConvertDMYToDays(D, M, Y, TempDays)
	then
		begin
			Day := D;
			Month := M;
			Year := Y;
			Days := TempDays;
		end;
end;

procedure TTime.SetDateTime(D, M, Y, H, Min, S: TSInt32);
begin
	SetDate(D, M, Y);
	SetTime(H, Min, S);
end;
{
procedure TTime.SetDateTimeFromString(D: String);
begin
end;
}
procedure TTime.SetMilliSeconds(M: TSInt32);
begin
	Millis := Abs(M mod 1000);
end;

procedure TTime.SetTime(H, M, S: TSInt32);
begin
	if (H >= 0) and (H <= 23)
		then Hour := H;
	if (M >= 0) and (M <= 59)
		then Minute := M;
	if (S >= 0) and (S <= 59)
		then Second := S;
end;

{ TTimeOptions }

constructor TTimeOptions.Init;
begin
	TimeString := '%w %d %n %Y %H:%L:%S';
	FirstDOW := 0; { Sunday }
	TimeZone := 0;
end;

begin
	CurrentTimeOptions.Init;
end.
